# Modified from https://github.com/fnpy/fn.py
# Refactor later

from .func import F, curry
from .array import Stream, Seq, AbstractSeq
from .underscore import shortcut as _
from .args import Args as __
from . import monad
from .monad import return_option, Option, Some, Empty

__all__ = ["F", "Stream", "_", "curry", "monad", "__", "return_option", "Option", "Some", "Empty", "AbstractSeq", "Seq"]



from typing import Union


class AverageMeter:
    """
    Computes and stores the average and current value.

    Attributes:
        val (float): The current value.
        avg (float): The average value.
        sum (float): The sum of all values.
        count (int): The number of values.
    """

    def __init__(self):
        self.count: int = 0
        self.sum: float = 0
        self.avg: float = 0
        self.val: float = 0

    def update(self, val: Union[int, float], n: int = 1) -> None:
        """
        Add new values to the average meter.

        Args:
            val(``int`` | ``float``): The new value to be added.
            n(``int``, optional): The number of values to be added.
        """
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count



from __future__ import annotations

from abc import ABC, abstractmethod


class AbstractServer(ABC):

    @abstractmethod
    def start(self):
        ...

    @abstractmethod
    def stop(self):
        ...

    def __enter__(self) -> AbstractServer:
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



from __future__ import annotations

from abc import ABC, abstractmethod
from functools import partial, wraps
from operator import eq, is_not
from typing import Callable, Generic, Union, Optional, Any, Tuple, List

from ...type import T, R, T1, T2, T3


class Option(ABC, Generic[T]):
    """
    ``Option`` represents optional values, each instance of ``Option``
    can be either instance of ``Some`` or ``Empty``.
    """
    empty: bool

    def __new__(cls, value: Union[T, Option[T]], checker: Callable[[T], bool] = partial(is_not, None)) -> Option[T]:
        if isinstance(value, Option):
            # Option(Some) -> Some
            # Option(Empty) -> Empty
            return value

        return Some(value) if checker(value) else Empty()

    def is_empty(self) -> bool:
        return self.empty

    def is_defined(self) -> bool:
        return not self.is_empty()

    @abstractmethod
    def flatten(self: Option[Option[T]]) -> Option[T]:
        ...

    @abstractmethod
    def map(self, f: Callable[[T], R]) -> Option[R]:
        ...

    @abstractmethod
    def flat_map(self, f: Callable[[T], Option[R]]) -> Option[R]:
        ...

    @abstractmethod
    def fold(self, f: Callable[[T], R], if_empty: R) -> R:
        ...

    @abstractmethod
    def for_each(self, f: Callable[[T], None]) -> None:
        ...

    @abstractmethod
    def filter(self, f: Callable[[T], bool]) -> Option[T]:
        ...

    @abstractmethod
    def filter_not(self, f: Callable[[T], bool]) -> Option[T]:
        ...

    @abstractmethod
    def exists(self, f: Callable[[T], bool]) -> bool:
        ...

    @abstractmethod
    def for_all(self, f: Callable[[T], bool]) -> bool:
        ...

    @abstractmethod
    def contains(self, value: T) -> bool:
        ...

    @abstractmethod
    def zip(self, other: Option[R]) -> Option[Tuple[T, R]]:
        ...

    @abstractmethod
    def unzip(self: Option[Tuple[T1, T2]]) -> Tuple[Option[T1], Option[T2]]:
        ...

    @abstractmethod
    def unzip3(self: Option[Tuple[T1, T2, T3]]) -> Tuple[Option[T1], Option[T2], Option[T3]]:
        ...

    @abstractmethod
    def get(self) -> Optional[T]:
        ...

    @abstractmethod
    def get_or_else(self, default: T) -> T:
        ...

    @abstractmethod
    def get_or_call(self, f: Callable[[...], T], *args, **kwargs) -> T:
        ...

    @abstractmethod
    def or_else(self, default: T) -> Option[T]:
        ...

    @abstractmethod
    def or_call(self, f: Callable[[...], T], *args, **kwargs) -> Option[T]:
        ...

    @abstractmethod
    def to_list(self) -> List[T]:
        ...


class Some(Option[T], Generic[T]):

    empty = False

    def __init__(self, value: T):
        # Option(Some) -> Some
        self.x: T = value.get() if isinstance(value, Some) else value

    def flatten(self: Some[Option[T]]) -> Option[T]:
        return self.x

    def map(self, f: Callable[[T], R]) -> Option[R]:
        return Option(f(self.x))

    def flat_map(self, f: Callable[[T], Option[R]]) -> Option[R]:
        return f(self.x)

    def fold(self, f: Callable[[T], R], if_empty: R) -> R:
        return f(self.x)

    def for_each(self, f: Callable[[T], None]) -> None:
        f(self.x)

    def filter(self, f: Callable[[T], bool]) -> Some[T]:
        return self if f(self.x) else Empty()

    def filter_not(self, f: Callable[[T], bool]) -> Option[T]:
        return self if not f(self.x) else Empty()

    def exists(self, f: Callable[[T], bool]) -> bool:
        return f(self.x)

    def for_all(self, f: Callable[[T], bool]) -> bool:
        return f(self.x)

    def contains(self, value: T) -> bool:
        return eq(self.x, value)

    def zip(self, other: Option[R]) -> Option[Tuple[T, R]]:
        if other.is_empty():
            return Empty()
        return Some((self.x, other.get()))

    def unzip(self: Some[Tuple[T1, T2]]) -> Tuple[Option[T1], Option[T2]]:
        if len(self.x) != 2 or not isinstance(self.x, tuple):
            raise ValueError("Pair tuple expected")
        a, b = self.x
        return Some(a), Some(b)

    def unzip3(self: Some[Tuple[T1, T2, T3]]) -> Tuple[Option[T1], Option[T2], Option[T3]]:
        a, b, c = self.x
        return Some(a), Some(b), Some(c)

    def get(self) -> T:
        return self.x

    def get_or_else(self, default: T) -> T:
        return self.x

    def get_or_call(self, f: Callable[[...], T], *args, **kwargs) -> T:
        return self.x

    def or_else(self, default: T) -> Some[T]:
        return self

    def or_call(self, f: Callable[[...], T], *args, **kwargs) -> Some[T]:
        return self

    def to_list(self) -> List[T]:
        return [self.x]

    def __str__(self) -> str:
        return "Some(%s)" % self.x

    __repr__ = __str__

    def __eq__(self, other: Option[Any]) -> bool:
        if not isinstance(other, Some):
            return False
        return eq(self.x, other.x)


class Empty(Option[None]):

    __object: Optional[Empty] = None  # singleton

    empty = True

    def __new__(cls) -> Empty:
        if Empty.__object is None:
            Empty.__object = object.__new__(cls)
        return Empty.__object

    def flatten(self: Empty) -> Empty:
        return Empty()

    def map(self, f: Callable[[T], R]) -> Empty:
        return Empty()

    def flat_map(self, f: Callable[[T], Option[R]]) -> Option[R]:
        return Empty()

    def fold(self, f: Callable[[T], R], if_empty: R) -> R:
        return if_empty

    def for_each(self, f: Callable[[T], None]) -> None:
        pass

    def filter(self, f: Callable[[T], bool]) -> Empty:
        return Empty()

    def filter_not(self, f: Callable[[T], bool]) -> Option[T]:
        return Empty()

    def exists(self, f: Callable[[T], bool]) -> bool:
        return False

    def for_all(self, f: Callable[[T], bool]) -> bool:
        return True

    def contains(self, value: T) -> bool:
        return False

    def zip(self, other: Option[R]) -> Empty:
        return Empty()

    def unzip(self: Empty) -> Tuple[Empty, Empty]:
        return Empty(), Empty()

    def unzip3(self: Empty) -> Tuple[Empty, Empty, Empty]:
        return Empty(), Empty(), Empty()

    def get_or_else(self, default: T) -> T:
        return default

    def get(self) -> None:
        return None

    def get_or_call(self, f: Callable[[...], T], *args, **kwargs) -> T:
        return f(*args, **kwargs)

    def or_else(self, default: T) -> Option[R]:
        return Option(default)

    def or_call(self, f: Callable[[...], T], *args, **kwargs) -> Option[T]:
        return Option(f(*args, **kwargs))

    def to_list(self) -> List:
        return []

    def __str__(self) -> str:
        return "Empty()"

    __repr__ = __str__

    def __eq__(self, other: Option[Any]) -> bool:
        return isinstance(other, Empty)


class ReturnOptionDecorator:
    """
    Decorator for modifying functions to return Option.
    It will return ``Empty`` if the return value doesn't meet checker.
    """

    def __init__(self, checker: Callable[[T], bool] = partial(is_not, None)):
        self._checker = checker

    def __call__(self, func: Callable[[...], T]) -> Callable[[...], Option[T]]:
        """
        Decorator for modifying functions to return Option.
        In default, it uses ``lambda x: x is not None`` as checker, when you use ``@return_option``.
        If you want to use your own checker, you can use ``@return_option.with_checker(checker)``.

        Args:
            func(``(...) -> T``): function to be modified to return Option[T].

        Returns:
            func(``(...) -> Option[T]``): modified function.

        Examples::

            from tensorneko.util import return_option

            @return_option
            def f1(x: int) -> int:
                return x + 1

            print(f1(1))  # Some(2)

            @return_option.with_checker(lambda x: x > 0)
            def f2(x: int) -> int:
                return x - 10

            print(f2(1))  # Empty()

        """
        @wraps(func)
        def wrapper(*args, **kwargs) -> Option[T]:
            return Option(func(*args, **kwargs), checker=self._checker)

        return wrapper

    @classmethod
    def with_checker(cls, checker: Callable[[T], bool]) -> Callable[[Callable[[...], T]], Callable[[...], Option[T]]]:
        return cls(checker)


return_option = ReturnOptionDecorator()



from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Executor, Future
from enum import Enum
from os import cpu_count
from typing import Optional, Dict, Union, Tuple

_num_workers = cpu_count() // 2


def set_num_workers(num_workers):
    global _num_workers
    _num_workers = num_workers


class ParallelType(Enum):
    THREAD = "thread"
    PROCESS = "process"


class ExecutorPool:
    thread_pools: Dict[str, ThreadPoolExecutor] = {}
    process_pools: Dict[str, ProcessPoolExecutor] = {}

    @classmethod
    def get_thread_executor(cls, name: Optional[str] = None, num_workers: int = _num_workers) -> ThreadPoolExecutor:
        name = name or "default"
        if name not in cls.thread_pools:
            cls.thread_pools[name] = ThreadPoolExecutor(num_workers)
        return cls.thread_pools[name]

    @classmethod
    def get_process_executor(cls, name: Optional[str] = None, num_workers: int = _num_workers) -> ProcessPoolExecutor:
        name = name or "default"
        if name not in cls.process_pools:
            cls.process_pools[name] = ProcessPoolExecutor(num_workers)
        return cls.process_pools[name]

    @classmethod
    def get(cls, name: Optional[str] = None, num_workers=_num_workers,
        parallel_type: ParallelType = ParallelType.PROCESS
    ) -> Executor:
        if parallel_type == ParallelType.THREAD:
            return cls.get_thread_executor(name, num_workers)
        elif parallel_type == ParallelType.PROCESS:
            return cls.get_process_executor(name, num_workers)
        else:
            raise ValueError("Invalid parallel type: {}".format(parallel_type))

    @classmethod
    def __class_getitem__(cls, item: Union[str, Tuple[str, ParallelType]]) -> Executor:
        if isinstance(item, str):
            if item in cls.thread_pools and item in cls.process_pools:
                raise ValueError("Ambiguous indexing. Both thread and process pool {} found".format(item))
            elif item in cls.thread_pools:
                return cls.thread_pools[item]
            elif item in cls.process_pools:
                return cls.process_pools[item]
            else:
                raise ValueError("Executor {} not found".format(item))
        elif isinstance(item, tuple):
            if len(item) != 2:
                raise ValueError("Invalid indexing. The format should be (name, parallel_type)")

            name, parallel_type = item
            if parallel_type == ParallelType.THREAD:
                return cls.get_thread_executor(name)
            elif parallel_type == ParallelType.PROCESS:
                return cls.get_process_executor(name)
            else:
                raise ValueError("Invalid parallel type: {}".format(parallel_type))
        else:
            raise ValueError("Invalid indexing. The format should be (name, parallel_type)")

    @classmethod
    def shutdown(cls, name: Optional[str] = None, parallel_type: Optional[ParallelType] = ParallelType.PROCESS) -> None:
        name = name or "default"
        if parallel_type == ParallelType.THREAD:
            if name in cls.thread_pools:
                cls.thread_pools[name].shutdown()
                del cls.thread_pools[name]
            else:
                raise ValueError("Thread pool {} not found".format(name))
        elif parallel_type == ParallelType.PROCESS:
            if name in cls.process_pools:
                cls.process_pools[name].shutdown()
                del cls.process_pools[name]
            else:
                raise ValueError("Process pool {} not found".format(name))
        else:
            raise ValueError("Invalid parallel type: {}".format(parallel_type))

    @classmethod
    def shutdown_all(cls):
        for name in cls.thread_pools:
            cls.thread_pools[name].shutdown()
        for name in cls.process_pools:
            cls.process_pools[name].shutdown()
        cls.thread_pools = {}
        cls.process_pools = {}

    @classmethod
    def submit(cls, func, *args, name: Optional[str] = None,
        parallel_type: Optional[ParallelType] = ParallelType.PROCESS, **kwargs
    ) -> Future:
        name = name or "default"

        return cls.get(name, parallel_type=parallel_type).submit(func, *args, **kwargs)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Union, Callable, List

from ...type import T, R


class AbstractSeq(ABC, Generic[T]):

    @abstractmethod
    def __getitem__(self, item: Union[int, slice]) -> Union[T, AbstractSeq[T]]:
        ...

    @abstractmethod
    def __iter__(self):
        ...

    @abstractmethod
    def map(self, f: Callable[[T], R]) -> AbstractSeq[R]:
        ...

    @abstractmethod
    def for_each(self, f: Callable[[T], None]) -> None:
        ...

    @abstractmethod
    def filter(self, f: Callable[[T], bool]) -> AbstractSeq[T]:
        ...

    @abstractmethod
    def reduce(self, f: Callable[[T, T], T]) -> T:
        ...

    @abstractmethod
    def flatten(self) -> AbstractSeq[T]:
        ...

    @abstractmethod
    def flat_map(self, f: Callable[[T], AbstractSeq[R]]) -> AbstractSeq[R]:
        ...

    @abstractmethod
    def skip(self, n: int) -> AbstractSeq[T]:
        ...

    @abstractmethod
    def take(self, n: int) -> AbstractSeq[T]:
        ...

    @abstractmethod
    def to_list(self) -> List[T]:
        ...



from __future__ import annotations

from typing import Union, Callable, Tuple, Any

from .func import F


class Args(F):
    """
    A variable wrapper for pipe operations.

    Args:
        args, kwargs: The arguments to be passed to the pipe.

    Examples::

        from tensorneko.util import __, _
        result = __(20) >> (_ + 1) >> (_ * 2) >> __.get
        print(result)
        # 42

    References:
        GitHub - fnpy/fp.py: Missing features of fp in Python -- active fork of kachayev/fp.py. (2022).
        Retrieved 5 April 2022, from https://github.com/fnpy/fn.py

        GitHub - kachayev/fp.py: Functional programming in Python: implementation of missing features to enjoy
        FP. (2022). Retrieved 5 April 2022, from https://github.com/kachayev/fn.py

    """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs

    def __ensure_callable(self, inputs):
        return Args(*(self.args + inputs), **self.kwargs) if isinstance(inputs, tuple) else inputs

    def __lshift__(self, g):
        """Overload << operator for Input and F instances"""
        raise ValueError("The Args instance cannot be the end of a pipe.")

    def __rshift__(self, g: Union[Callable, Tuple, Args, F]) -> Union[Args, any]:
        """Overload >> operator for F instances"""
        if type(g) is Args:
            self.kwargs.update(g.kwargs)
            return Args(*(self.args + g.args), **self.kwargs)
        elif isinstance(g, tuple):
            return self.__ensure_callable(g)
        elif any(map(lambda getter: g is getter, (Args.get, Args.get_args, Args.get_value, Args.get_kwargs))):
            return g(self)
        else:
            return Args(g(*self.args, **self.kwargs))

    def __call__(self, *args, **kwargs):
        """Overload apply operator"""
        raise TypeError("The 'Args' is not callable")

    def __str__(self):
        kwargs_str = ""
        for k, v in self.kwargs.items():
            kwargs_str = f"{kwargs_str}, {k}={v}"

        return f"({', '.join(map(str, self.args))}{kwargs_str})"

    __repr__ = __str__

    def get(self) -> Any:
        return self.args[0]

    def get_args(self) -> tuple:
        return self.args

    def get_kwargs(self) -> dict:
        return self.kwargs

    def get_value(self, key: str) -> dict:
        return self.kwargs[key]



from __future__ import annotations

from enum import Enum
from functools import reduce
from itertools import chain, islice
from sys import maxsize
from typing import Iterable, Iterator, List, Union, Callable, Optional

from .abstract_seq import AbstractSeq
from ...type import T, R


class Stream(AbstractSeq, Iterable[T]):

    def __init__(self, iterable: Iterable[T] = iter(()), action_pipe: List[_StreamAction] = None):
        if isinstance(iterable, Stream):
            self._iter = iterable._iter
            self._action_pipe = action_pipe or iterable._action_pipe
            self._cache = iterable._cache
            self._current_index = iterable._current_index
        else:
            self._iter = iter(iterable)
            self._action_pipe = action_pipe or []
            self._cache = []
            self._current_index = -1

    @staticmethod
    def of(*items: T) -> Stream[T]:
        return Stream(items)

    @staticmethod
    def from_stream(stream: Stream[T]) -> Stream[T]:
        return Stream(stream._iter)

    def __lshift__(self, right_iter: Iterable[T]) -> Stream[T]:
        self._iter = chain(self._iter, right_iter)
        return self

    def __iter__(self) -> _Iterator:
        return self._Iterator(self)

    class _Iterator(Iterator[T]):

        def __init__(self, stream: Stream[T]):
            self._stream = stream
            self._iter_index = -1

        def __next__(self) -> T:
            self._iter_index += 1
            result = self._stream._try_iter_to(self._iter_index)
            if result in (self._stream._IterStatus.SUCCESS_ITER, self._stream._IterStatus.NO_NEED_ITER):
                return self._stream._cache[self._iter_index]
            else:
                raise StopIteration

    class _IterStatus(Enum):
        FAIL = 0
        NO_NEED_ITER = 1
        SUCCESS_ITER = 2

    def _try_iter_to(self, index: int) -> _IterStatus:
        if index <= self._current_index:
            return self._IterStatus.NO_NEED_ITER

        while index > self._current_index:
            try:
                item = next(self._iter)
            except StopIteration:
                return self._IterStatus.FAIL

            filtered = True
            flattened = False
            for i, (f, method) in enumerate(self._action_pipe):
                if method == Stream.map:
                    item = f(item)
                elif method == Stream.filter:
                    filtered = f(item)
                    if not filtered:
                        break
                elif method == Stream.flatten:
                    if isinstance(item, Stream):
                        item = Stream(item, action_pipe=self._action_pipe[i + 1:]).to_list()
                        flattened = True
                        break  # the remain actions are already applied in the sub stream to_list

            if filtered:
                if flattened:
                    self._cache.extend(item)
                    self._current_index += len(item)
                else:
                    self._cache.append(item)
                    self._current_index += 1

        return self._IterStatus.SUCCESS_ITER

    def __getitem__(self, item: Union[int, slice]) -> Union[T, Stream[T]]:
        if isinstance(item, int):
            if item < 0:
                raise ValueError("Negative slice is not allowed in Stream")
            self._try_iter_to(item)
        elif isinstance(item, slice):
            low, high, step = item.indices(maxsize)
            if step == 0:
                raise ValueError("Step must not be 0")
            return self.__class__() << map(self.__getitem__, range(low, high, step or 1))
        else:
            raise TypeError("Invalid argument type")

        return self._cache.__getitem__(item)

    def map(self, f: Callable[[T], R]) -> Stream[R]:
        return Stream(self._iter, action_pipe=self._action_pipe + [_StreamAction(f, Stream.map)])

    def for_each(self, f: Callable[[T], None]) -> None:
        for each in self.to_list():
            f(each)

    def filter(self, f: Callable[[T], bool]) -> Stream[T]:
        return Stream(self._iter, action_pipe=self._action_pipe + [_StreamAction(f, Stream.filter)])

    def reduce(self, f: Callable[[T, T], T]) -> T:
        return reduce(f, self)

    def flatten(self) -> Stream[T]:
        return Stream(self._iter, action_pipe=self._action_pipe + [_StreamAction(None, Stream.flatten)])

    def flat_map(self, f: Callable[[T], AbstractSeq[R]]) -> AbstractSeq[R]:
        return self.map(f).flatten()

    def skip(self, n: int) -> Stream[T]:
        return Stream(islice(self, n, None))

    def take(self, n: int) -> Stream[T]:
        return Stream(islice(self, n))

    def to_list(self) -> List[T]:
        return [*self]

    def __str__(self):
        return f"Stream({', '.join(map(str, self._cache + ['...']))})"


class _StreamAction:

    def __init__(self, f: Optional[Callable], method: Callable):
        self.f = f
        self.method = method

    def __iter__(self):
        return iter((self.f, self.method))

    def __str__(self):
        return f"{self.method.__name__}({self.f.__name__ if self.f is not None else ''})"
